/****************************************************************************
* MeshLab                                                           o o     *
* A versatile mesh processing toolbox                             o     o   *
*                                                                _   O  _   *
* Copyright(C) 2005                                                \/)\/    *
* Visual Computing Lab                                            /\/|      *
* ISTI - Italian National Research Council                           |      *
*                                                                    \      *
* All rights reserved.                                                      *
*                                                                           *
* This program is free software; you can redistribute it and/or modify      *   
* it under the terms of the GNU General Public License as published by      *
* the Free Software Foundation; either version 2 of the License, or         *
* (at your option) any later version.                                       *
*                                                                           *
* This program is distributed in the hope that it will be useful,           *
* but WITHOUT ANY WARRANTY; without even the implied warranty of            *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
* GNU General Public License (http://www.gnu.org/licenses/gpl.txt)          *
* for more details.                                                         *
*                                                                           *
****************************************************************************/

#include "filter_manifold.h"
#include <QtScript>

using namespace vcg;

int CountIncidentFaces(CMeshO::FacePointer fp, int edgeInd)
{
  CMeshO::FacePointer fpStart = fp;
  CMeshO::FacePointer fpCur = fp;
  int eiCur=edgeInd;
  int incidentCnt=0;
  do
  {
    CMeshO::FacePointer fpNext=fpCur->FFp(eiCur);
    int eiNext=fpCur->FFi(eiCur);
    incidentCnt++;
    eiCur=eiNext;
    fpCur=fpNext;
  }
  while(fpCur!=fpStart);

  return incidentCnt;
}

bool SelectMinimalInFaceRing(CMeshO::FacePointer fp, int edgeInd)
{
  CMeshO::FacePointer fpStart = fp;
  CMeshO::FacePointer fpCur = fp;
  CMeshO::FacePointer fpBest = fp;
  int eiCur=edgeInd;

  float minArea=std::numeric_limits<float>::max();
  do
  {
    CMeshO::FacePointer fpNext=fpCur->FFp(eiCur);
    int eiNext=fpCur->FFi(eiCur);

    if(DoubleArea(*fpCur) < minArea)
    {
      minArea=DoubleArea(*fpCur);
      fpBest=fpCur;
    }

    eiCur=eiNext;
    fpCur=fpNext;
  }
  while(fpCur!=fpStart);
  fpBest->SetS();
  return true;
}


// Constructor usually performs only two simple tasks of filling the two lists 
//  - typeList: with all the possible id of the filtering actions
//  - actionList with the corresponding actions. If you want to add icons to your filtering actions you can do here by construction the QActions accordingly

FilterManifoldPlugin::FilterManifoldPlugin()
{ 
  typeList << FP_SMART_REMOVE_MANIFOLD;
  
  foreach(FilterIDType tt , types())
	  actionList << new QAction(filterName(tt), this);
}

// ST() must return the very short string describing each filtering action 
// (this string is used also to define the menu entry)
QString FilterManifoldPlugin::filterName(FilterIDType filterId) const
{
  switch(filterId) {
    case FP_SMART_REMOVE_MANIFOLD :  return QString("Smart non Manifold Selection");
		default : assert(0); 
	}
  return QString();
}

// Info() must return the longer string describing each filtering action 
// (this string is used in the About plugin dialog)
 QString FilterManifoldPlugin::filterInfo(FilterIDType filterId) const
{
  switch(filterId) {
    case FP_SMART_REMOVE_MANIFOLD :  return QString("Select non Manifold faces choosing etc etc");
		default : assert(0); 
	}
	return QString("Unknown Filter");
}

// The FilterClass describes in which generic class of filters it fits. 
// This choice affect the submenu in which each filter will be placed 
// More than a single class can be choosen.
FilterManifoldPlugin::FilterClass FilterManifoldPlugin::getClass(QAction *a)
{
  switch(ID(a))
	{
  case FP_SMART_REMOVE_MANIFOLD :  return MeshFilterInterface::Cleaning;
		default : assert(0); 
	}
	return MeshFilterInterface::Generic;
}

// This function define the needed parameters for each filter. Return true if the filter has some parameters
// it is called every time, so you can set the default value of parameters according to the mesh
// For each parameter you need to define, 
// - the name of the parameter, 
// - the string shown in the dialog 
// - the default value
// - a possibly long string describing the meaning of that parameter (shown as a popup help in the dialog)
void FilterManifoldPlugin::initParameterSet(QAction *action,MeshModel &m, RichParameterSet & parlst)
{
  QStringList methods;
   switch(ID(action))	 {
    case FP_SMART_REMOVE_MANIFOLD :
     // two possibilities
     //- quella che ha il maggior numero di edge non manifold o bordi
     // - quella di area minima
         methods.push_back("Minimal Area");
         methods.push_back("Less Connected");
         parlst.addParam(new RichEnum("Removalstrategy", 0, methods, tr("Removal strategy"), tr("Choose what strategy has to be applied to transform each non manifold edge in a 2 manifold one")));
											break;
											
		default : assert(0); 
	}
}

// The Real Core Function doing the actual mesh processing.
// Move Vertex of a random quantity
bool FilterManifoldPlugin::applyFilter(QAction */*filter*/, MeshDocument &md, RichParameterSet & par, vcg::CallBackPos *cb)
{
  MeshModel &m=*md.mm();

  m.updateDataMask(MeshModel::MM_FACEFACETOPO);

  tri::UpdateFlags<CMeshO>::FaceClearV(m.cm);
  tri::UpdateNormals<CMeshO>::PerVertexNormalized(m.cm);
  int selCnt=0;
  CMeshO::FaceIterator fi;
  for(fi=m.cm.face.begin();fi!=m.cm.face.end();++fi)
  {
    for(int i=0;i<3;++i)
    {
      // To set a face as visited
      // (*fi).SetV();
      // To test a face if it is visited
      // (*fi).IsV();
      // (*fi).ClearV();
      int nonManifDegree = CountIncidentFaces(&*fi,i);
      if(nonManifDegree>2)
      {
        SelectMinimalInFaceRing(&*fi,i);
        selCnt++;
      }
    }
  }
    // Log function dump textual info in the lower part of the MeshLab screen.
  Log("Selected %i faces",selCnt);
	
  vcg::tri::UpdateBounding<CMeshO>::Box(m.cm);
  
	return true;
}

QString FilterManifoldPlugin::filterScriptFunctionName( FilterIDType filterID )
{
	switch(filterID) {
    case FP_SMART_REMOVE_MANIFOLD :  return QString("randomVerticesDisplacement");
		default : assert(0); 
	}
	return QString();
}

Q_EXPORT_PLUGIN(FilterManifoldPlugin)
